/******************************************************************************
 *                                                                            *
 * File:   regread.c          Version  1.11            Date: 2000-06-14       *
 *                                                                            *
 * Copyright (C) 2000 by Kostis Netzwerkberatung                              *
 * Talstr. 25, D-63322 Rdermark, Tel. +49 6074 881056, FAX 881058            *
 * kosta@kostis.net (Kosta Kostis), http://www.kostis.net/                    *
 *                                                                            *
 ******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *        - displays values from regstry                                      *
 *                                                                            *
 ******************************************************************************
 *                                                                            *
 * History:                                                                   *
 *     2000-06-14: KK V1.11                                                   *
 *        - fixed bug for first registry level (such as Software)             *
 *     2000-06-03: KK V1.10                                                   *
 *        - use commas instead of tabs in query files (added level)           *
 *        - moved syntax check into external function ScanFlag ()             *
 *        - allow for recoursive queries                                      *
 *     2000-01-21: KK V1.02                                                   *
 *        - more comments, cosmetics, more clear debugging messages           *
 *     2000-01-20: KK V1.01                                                   *
 *        - debugged with some help of Rolf.Zinn@igmetall.de                  *
 *     2000-01-19: KK V1.00                                                   *
 *        - initial encoding                                                  *
 *                                                                            *
 *****************************************************************************/

#include <windows.h>
#include <winbase.h>
#include <winreg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "scanflag.h"
#include "regread.h"

/******************************************************************************

	Function:
		display copyright, author and program information

	Parameters:
		none

	Returns:
		nothing

 *****************************************************************************/

static  void    Hello (void)
{
	printf (MSG_COPYRIGHT) ;
	printf (MSG_VERSION, PRGNAM, VERSION) ;
}

/******************************************************************************

	Function:
		parse key root

	Parameters:
		char	*RegKey

	Returns:
		HKEY_DYN_DATA		if key is unknown

		HKEY_CLASSES_ROOT
	or	HKEY_CURRENT_CONFIG
	or	HKEY_CURRENT_USER
	or	HKEY_LOCAL_MACHINE
	or	HKEY_USERS

 *****************************************************************************/

static	HKEY	ParseKey
(
	char	*RegRoot
)
{
	HKEY	hkType ;

	/* determine HKEY (crude but effective) */
	hkType = HKEY_DYN_DATA ;
	if (strcmp (RegRoot, STR_HKEY_CLASSES_ROOT) == 0)
		hkType = HKEY_CLASSES_ROOT ;
	if (strcmp (RegRoot, STR_HKEY_CURRENT_CONFIG) == 0)
		hkType = HKEY_CURRENT_CONFIG ;
	if (strcmp (RegRoot, STR_HKEY_CURRENT_USER) == 0)
		hkType = HKEY_CURRENT_USER ;
	if (strcmp (RegRoot, STR_HKEY_LOCAL_MACHINE) == 0)
		hkType = HKEY_LOCAL_MACHINE ;
	if (strcmp (RegRoot, STR_HKEY_USERS) == 0)
		hkType = HKEY_USERS ;

	return (hkType) ;
}

/******************************************************************************

	Function:
		if FileName is not given, create default FileName

	Parameters:
		char	*FileName	does/will contain FileName
		char	*argv0		program name (including path)

	Returns:
		nothing

	If FileName is not given, FileName will be comprised of the program
	name but with the extension ".rqu" and not ".exe"

 *****************************************************************************/

void	GetFileName
(
	char	*FileName,
	char	*argv0
)
{
	char	*Help ;

	if (strlen (FileName) > 0)
		return ;

	/* search query file in program directory */
	if (DEBUG)
		printf ("%s: argv0=%s\n", PRGNAM, argv0) ;

	/* make a copy of the program file name */
	strcpy (FileName, argv0) ;

	/* look for last "." (as in ".exe") */
	Help = strrchr (FileName, '.') ;

	/* cut off ".exe" if present */
	if (Help != (char *) NULL)
		*Help = '\0' ;

	/* add EXT_QUERYFILE to filename */
	strcat (FileName, EXT_QUERYFILE) ;
}

/******************************************************************************

	Function:
		dump registry keys below a given RegKey

	Parameters:
		HKEY		hKey
		char		*RegKey
		int		Level

	Returns:
		nothing (just displays results using enumeration)

 *****************************************************************************/

static	void RegDump
(
	char		*RegPrefix,
	HKEY		hkType,
	HKEY		hKey,
	char		*RegKey,
	int		Level
)
{
	HKEY		hSubkey ;
	char		RegNewkey	[MAX_LEN_REG_VALUE + 1] ;
	char		RegValue	[MAX_LEN_REG_VALUE + 1] ;
	char		RegResult	[MAX_LEN_REG_VALUE + 1] ;
	DWORD		dwIndex ;
	DWORD		dwBytes ;
	DWORD		dwType ;
	DWORD		dwData ;
	FILETIME	ftLastChanged ;
	LONG		rc ;

	dwIndex = 0 ;
	dwBytes = MAX_LEN_REG_VALUE ;
	dwData = MAX_LEN_REG_VALUE ;

	if (DEBUG)
		printf ("%s: RegDump (%s, hKey, %s, %d) ;\n",
			PRGNAM, RegPrefix, RegKey, Level) ;

	/**********************************************************************

		list all values

	 *********************************************************************/

	while
	(
		RegEnumValue
		(
			hKey, dwIndex++,
			RegValue, &dwBytes,
			NULL, &dwType,
			RegResult, &dwData
		) == ERROR_SUCCESS
	)
	{
		if (DEBUG)
			printf ("%s: RegEnumValue () found %s\\%s\n",
				PRGNAM, RegKey, RegValue) ;

		/**************************************************************

			display value if it's a simple string

		 *************************************************************/

		switch (dwType)
		{
			case	REG_SZ:

				printf ("%s,\"%s\\%s\",\"%s\"\n",
					RegPrefix, RegKey,
					RegValue, RegResult) ;
				break ;

			case	REG_DWORD:

				printf ("%s,\"%s\\%s\",\"%d\"\n",
					RegPrefix, RegKey,
					RegValue, (DWORD) *RegResult) ;
				break ;
		}

		/**************************************************************

			reset variables for next call of RegEnumValue()

		 *************************************************************/

		dwBytes = MAX_LEN_REG_VALUE ;
		dwData = MAX_LEN_REG_VALUE ;
	}

	if (Level > 0)
		--Level ;

	/**********************************************************************

		list all subkeys (if there are any)

	 *********************************************************************/

	dwIndex = 0 ;
	dwBytes = MAX_LEN_REG_VALUE ;
	while
	(
		(
			rc = RegEnumKeyEx
			(
				hKey, dwIndex++,
				RegValue, &dwBytes,
				NULL, NULL, NULL, &ftLastChanged
			)
		) == ERROR_SUCCESS
	)
	{
		/**************************************************************

			display current subkey

		 *************************************************************/

		sprintf (RegNewkey, "%s\\%s", RegKey, RegValue) ;
		printf ("%s,\"%s\",\"...\"\n", RegPrefix, RegNewkey) ;

		/**************************************************************

			if recursion level permits, investigate subkey

		 *************************************************************/

		if (Level)
		{
			rc = RegOpenKeyEx (hkType, RegNewkey, 0,
				KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
				&hSubkey) ;
			RegDump (RegPrefix, hkType, hSubkey, RegNewkey, Level) ;
			RegCloseKey (hSubkey) ;
		}

		/**************************************************************

			reset variable for next call of RegEnumKeyEx()

		 *************************************************************/

		dwBytes = MAX_LEN_REG_VALUE ;
	}

	return ;
}

/******************************************************************************

	main program

 *****************************************************************************/

int     main
(
	int     	argc,
	char		*argv	[]
)
{
	/* variables for computername */
	char		Computername	[MAX_LEN_COMPUTERNAME + 1] ;
	DWORD		cnlen		= MAX_LEN_COMPUTERNAME ;

	/* variables for queries */
	char		*Help ;
	char		*Next ;
	int		Level ;

	/* variables for reading query file */
	FILE		*Fin ;
	char		Line		[MAX_LEN_REG_VALUE + 1] ;
	char		RegRoot		[MAX_LEN_REG_VALUE + 1] ;
	char		RegKey		[MAX_LEN_REG_VALUE + 1] ;
	char		RegSubkey	[MAX_LEN_REG_VALUE + 1] ;
	char		RegValue	[MAX_LEN_REG_VALUE + 1] ;
	char		RegResult	[MAX_LEN_REG_VALUE + 1] ;
	char		RegPrefix	[MAX_LEN_REG_VALUE + 1] ;

	/* variables for Reg* () */
	HKEY		hkType ;
	HKEY		hkResult ;
	HKEY		hkRecoursion ;
	DWORD		dwType ;
	DWORD		dwData ;
	LONG		rc ;

	/**********************************************************************

		parse parameters/arguments

	 *********************************************************************/

	strcpy (FileName, "") ;
	if (ScanFlag (argc, argv) != SCANFLAG_OK)
	{
		/**************************************************************

			display syntax error message and quit

		 *************************************************************/

		Hello () ;
		fprintf (stderr, ERR_MSG_SYNTAX, PRGNAM) ;

		return (ERR_SYNTAX) ;
	}

	if (DEBUG)
		Hello () ;

	/**********************************************************************

		get computername

	 *********************************************************************/

	GetComputerName (Computername, &cnlen) ;
	if (DEBUG)
		printf ("%s: GetComputerName()=%s\n", PRGNAM, Computername) ;

	/**********************************************************************

		if no filename is given use default filename

	 *********************************************************************/

	GetFileName (FileName, argv [0]) ;
	if (DEBUG)
		printf ("%s: GetFileName()=%s\n",
			PRGNAM, FileName) ;


	/**********************************************************************

		open registry query file

	 *********************************************************************/

	Fin = fopen (FileName, "r") ;
	if (Fin == (FILE *) NULL)
	{
		if (DEBUG)
			printf ("%s: unable to open query file %s\n",
				PRGNAM, FileName) ;
		return (ERR_NOQUERYFILE) ;
	}
	if (DEBUG)
		printf ("%s: opened query file %s\n",
			PRGNAM, FileName) ;

	/**********************************************************************

		process registry query file line by line

	 *********************************************************************/

	for (; feof (Fin) == 0 ;)
	{
		/**************************************************************

			read line containing HKEY,Level,RegKey

		 *************************************************************/

		strcpy (Line, "") ;
		fgets (Line, MAX_LEN_REG_VALUE, Fin) ;

		/**************************************************************

			check for end of file

		 *************************************************************/

		if (feof (Fin) != 0)
		{
			if (DEBUG)
				printf ("%s: end of file\n", PRGNAM) ;
			fclose (Fin) ;
			return (ERR_NONE) ;
		}

		/**************************************************************

			remove trailing '\n' (coming from fgets ())

		 *************************************************************/

		Help = strchr (Line, '\n') ;
		if (Help != (char *) NULL)
			*Help = '\0' ;
		if (DEBUG)
			printf ("%s: read line \"%s\"\n", PRGNAM, Line) ;

		/**************************************************************

			skip empty lines

		 *************************************************************/

		if (strlen (Line) == 0)
		{
			if (DEBUG)
				printf ("%s: skipping empty line\n", PRGNAM) ;
			continue ;
		}

		/**************************************************************

			find position for line split at ','

		 *************************************************************/

		Help = strchr (Line, ',') ;
		if (Help == (char *) NULL)
		{
			if (DEBUG)
				printf ("%s: ',' 1 not found - skipping line.\n",
					PRGNAM) ;
			continue ;
		}

		/**************************************************************

			split string, copy HKEY, Help points to level

		 *************************************************************/

		*Help++ = '\0' ;
		strcpy (RegRoot, Line) ;

		/**************************************************************

			determine HKEY

		 *************************************************************/

		CharUpper (RegRoot) ;	/* Parse UPPER CASE version */
		hkType = ParseKey (RegRoot) ;
		if (hkType == HKEY_DYN_DATA)
		{
			/* HKEY unknown */
			if (DEBUG)
				printf ("%s: HKEY %s unknown\n",
					PRGNAM, RegRoot) ;
			continue ;
		}

		/**************************************************************

			find next field

		 *************************************************************/

		Next = strchr (Help, ',') ;
		if (Next == (char *) NULL)
		{
			if (DEBUG)
				printf ("%s: ',' 2 not found - skipping line.\n",
					PRGNAM) ;
			continue ;
		}

		/**************************************************************

			retrieve recoursion level (-1 means unlimited)

		 *************************************************************/

		*Next++ = '\0' ;
		Level = atoi (Help) ;
		strcpy (RegKey, Next) ;

		/**************************************************************

			store computername, HKEY and registry key

		 *************************************************************/

		sprintf (RegPrefix, "\"%s\",\"%s\"", Computername, RegRoot) ;

		/**************************************************************

			determine value name (after last backslash)

		 *************************************************************/

		strcpy (RegSubkey, RegKey) ;
		Help = strrchr (RegSubkey, '\\') ;
		if (Help != (char *) NULL)
		{
			*Help++ = '\0' ;
			strcpy (RegValue, Help) ;
		}
		else
		{
			strcpy (RegValue, "") ;
		}

		if (DEBUG)
			printf ("%s: HKEY=%s, Level=%d, RegSubkey=%s, RegValue=%s\n",
				PRGNAM, RegRoot, Level, RegSubkey, RegValue) ;

		/**************************************************************

			set defaults for Reg*() calls

		 *************************************************************/

		dwData = MAX_LEN_REG_VALUE ;
		strcpy (RegResult, "") ;

		/**************************************************************

			try to open registry key

		 *************************************************************/

		rc = RegOpenKeyEx (hkType, RegSubkey, 0,
			KEY_QUERY_VALUE, &hkResult) ;
		if (rc == 0)
		{
			/******************************************************

				try to query value

			 *****************************************************/

			rc = 42 ;
			if (strlen (RegValue) > 0)
				rc = RegQueryValueEx (hkResult, RegValue, 0,
					&dwType, RegResult, &dwData) ;
			if (rc == 0)
			{
				printf ("%s,\"%s\",\"%s\"\n",
					RegPrefix, RegKey, RegResult) ;
			}
			else
			{
				/**********************************************

					check if it can be opened as a key

				 *********************************************/

				rc = RegOpenKeyEx (hkType, RegKey, 0,
					KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS,
					&hkRecoursion) ;
				if (rc)
				{
					printf ("%s,\"%s\",\"-\"\n",
						RegPrefix, RegKey) ;
					if (DEBUG)
						printf ("%s: RegQueryValueEx () returned %d\n",
							PRGNAM, rc) ;
				}
				else
				{
					/**************************************

						we need recoursion

					 *************************************/

					if (DEBUG)
						printf ("%s: recursion for %s\n",
							PRGNAM, RegKey) ;
					RegDump (RegPrefix, hkType,
						hkRecoursion, RegKey, Level) ;
				}
				RegCloseKey (hkRecoursion) ;
			}
			RegCloseKey (hkResult) ;
		}
		else
		{
			printf ("%s,\"%s\",\"-\"\n", RegPrefix, RegKey) ;
			if (DEBUG)
				printf ("%s: RegOpenKeyEx () returned %d\n",
					PRGNAM, rc) ;
		}
		
	}

	/**********************************************************************

		we encountered EOF before '\n'

	 *********************************************************************/
	
	if (DEBUG)
		printf ("%s: end of file\n", PRGNAM) ;
	fclose (Fin) ;

	return (ERR_NONE) ;
}
